@@ -729,12 +729,11 @@ class ProcessError(Error):
 
         if exit_signal:
             signal, core_dumped, msg, lang = exit_signal
-            reason = 'Process exited with signal %s%s%s' % \
-                (signal, ': ' + msg if msg else '',
-                 ' (core dumped)' if core_dumped else '')
+            reason = 'Process exited with signal ' + signal + \
+                     (': ' + msg if msg else '') + \
+                     (' (core dumped)' if core_dumped else '')
         elif exit_status:
-            reason = 'Process exited with non-zero exit status %s' % \
-                exit_status
+            reason = f'Process exited with non-zero exit status {exit_status}'
 
         super().__init__(exit_status or 0, reason, lang)
 
@@ -929,7 +928,7 @@ class SSHProcess(SSHStreamSession, Generic[AnyStr]):
                 file = source
 
             if hasattr(file, 'read') and \
-                    (asyncio.iscoroutinefunction(file.read) or
+                    (inspect.iscoroutinefunction(file.read) or
                      inspect.isgeneratorfunction(file.read)):
                 reader = _AsyncFileReader(self, cast(_AsyncFileProtocol, file),
                                           bufsize, datatype, self._encoding,
@@ -998,7 +997,7 @@ class SSHProcess(SSHStreamSession, Generic[AnyStr]):
                 needs_close = recv_eof
 
             if hasattr(file, 'write') and \
-                    (asyncio.iscoroutinefunction(file.write) or
+                    (inspect.iscoroutinefunction(file.write) or
                      inspect.isgeneratorfunction(file.write)):
                 writer = _AsyncFileWriter(
                     self, cast(_AsyncFileProtocol, file), needs_close,
@@ -1211,6 +1210,8 @@ class SSHProcess(SSHStreamSession, Generic[AnyStr]):
         for task in self._cleanup_tasks:
             await task
 
+        self._cleanup_tasks = []
+
 
 class SSHClientProcess(SSHProcess[AnyStr], SSHClientStreamSession[AnyStr]):
     """SSH client process handler"""
