@@ -25,6 +25,7 @@ import binascii
 import functools
 import os
 import shutil
+import socket
 import subprocess
 import sys
 import tempfile
@@ -93,6 +94,31 @@ def asynctest(coro):
     return async_wrapper
 
 
+def patch_getaddrinfo(cls):
+    """Decorator for patching socket.getaddrinfo"""
+
+    # pylint: disable=redefined-builtin
+
+    cls.orig_getaddrinfo = socket.getaddrinfo
+
+    hosts = {'testhost.test': '',
+             'testcname.test': 'cname.test',
+             'cname.test': ''}
+
+    def getaddrinfo(host, port, family=0, type=0, proto=0, flags=0):
+        """Mock DNS lookup of server hostname"""
+
+        # pylint: disable=unused-argument
+
+        try:
+            return [(socket.AF_INET, socket.SOCK_STREAM, socket.IPPROTO_TCP,
+                     hosts[host], ('127.0.0.1', port))]
+        except KeyError:
+            return cls.orig_getaddrinfo(host, port, family, type, proto, flags)
+
+    return patch('socket.getaddrinfo', getaddrinfo)(cls)
+
+
 def patch_getnameinfo(cls):
     """Decorator for patching socket.getnameinfo"""
 
@@ -106,6 +132,19 @@ def patch_getnameinfo(cls):
     return patch('socket.getnameinfo', getnameinfo)(cls)
 
 
+def patch_getnameinfo_error(cls):
+    """Decorator for patching socket.getnameinfo to raise an error"""
+
+    def getnameinfo_error(sockaddr, flags):
+        """Mock failure of reverse DNS lookup of client address"""
+
+        # pylint: disable=unused-argument
+
+        raise socket.gaierror()
+
+    return patch('socket.getnameinfo', getnameinfo_error)(cls)
+
+
 def patch_extra_kex(cls):
     """Decorator for skipping extra kex algs"""
 
@@ -202,7 +241,7 @@ def make_certificate(cert_version, cert_type, key, signing_key, principals,
     """Construct an SSH certificate"""
 
     keydata = key.encode_ssh_public()
-    principals = b''.join((String(p) for p in principals))
+    principals = b''.join(String(p) for p in principals)
     options = _encode_options(options) if options else b''
     extensions = _encode_options(extensions) if extensions else b''
     signing_keydata = b''.join((String(signing_key.algorithm),
@@ -309,7 +348,7 @@ class ConnectionStub:
                 self.connection_lost(data)
                 break
 
-            self.process_packet(data)
+            await self.process_packet(data)
 
     def connection_lost(self, exc):
         """Handle the closing of a connection"""
@@ -409,8 +448,6 @@ class AsyncTestCase(TempDirTestCase):
         else:
             cls.loop = asyncio.new_event_loop()
 
-        asyncio.set_event_loop(cls.loop)
-
         try:
             cls.loop.run_until_complete(cls.asyncSetUpClass())
         except AttributeError:
